#! /usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import sys
import os
import re
import shutil
import collections
import subprocess
import dendropy
from dendropy.utility import processio
from archipelago import model

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import study

class Processor(object):

    def __init__(
            self,
            path_to_master_reference,
            path_to_master_tree,
            ):
        self.system = study.SystemConfiguration(
            path_to_master_reference=path_to_master_reference,
            path_to_master_tree=path_to_master_tree,
            )

    def compose_subpath(self,
            subdir_paths,
            file_subname):
        if subdir_paths is None:
            subdir_paths = []
        args = subdir_paths + [self.output_name + "." + file_subname]
        return os.path.join(self.output_dirpath, *args)

    def ensure_project_subdirectory(self, dirpath):
        if not os.path.exists(dirpath):
            os.makedirs(dirpath)

    def clean_directory(self, *dirpaths):
        dirpath = os.path.join(dirpaths)
        filepaths = os.listdir(dirpath)
        for filepath in filepaths:
            filepath = os.path.join(dirpath, filepath)
            os.remove(filepath)

    def process(self):
        for spatial_regime in ("island", "module"):
            root_dir = os.path.join("reference", "data", "by_{}".format(spatial_regime), "trees")
            self.ensure_project_subdirectory(root_dir)

            for trait_name, lineage_trait_map in (
                    ("habitat-disturbed-vs-undisturbed", self.system.lineage_name_disturbed_vs_undisturbed_habitat_map),
                    ("habitat-generalist-vs-interior", self.system.lineage_name_generalist_vs_interior_habitat_map),
                    ("role-widespread-vs-localized", self.system.lineage_name_widespread_vs_localized_role_map),
                    ("trophic-level", self.system.lineage_name_trophic_level_map),
                    ):
                additional_lineage_names_to_drop = [lineage_name for lineage_name in lineage_trait_map if lineage_trait_map[lineage_name] is None]
                tree = self.system.get_new_master_tree(
                        restrict_to_study_lineages=True,
                        additional_lineage_names_to_drop=additional_lineage_names_to_drop,
                        distribution_vector_type=spatial_regime)
                trait_states = set()
                for tidx, taxon in enumerate(tree.taxon_namespace):
                    taxon.index = tidx + 1
                    trait_state = lineage_trait_map[taxon.label]
                    taxon.traits_vector = [trait_state]
                    trait_states.add(trait_state)
                    taxon.label = model.ArchipelagoModel.compose_encoded_label(
                            lineage=taxon,
                            excluded_area_indexes=None)
                trait_states = sorted(trait_states)
                print("By {}, {}: {} taxa, {} areas, {} trait states ({})".format(
                    spatial_regime,
                    trait_name,
                    len(tree.taxon_namespace),
                    tree.num_areas,
                    len(trait_states),
                    trait_states,
                    ))
                with open(os.path.join(root_dir, "wallacea.by-{}.{}.newick".format(spatial_regime, trait_name)), "w") as out:
                    tree.write_to_stream(
                            out,
                            "newick",
                            suppress_internal_node_labels=True,
                            suppress_internal_taxon_labels=True,
                            )

processor = Processor(
        path_to_master_reference="reference/sources/master_reference.csv",
        path_to_master_tree="reference/sources/master_tree.nexus",
        )
processor.process()
# for f in trees/*.newick ; do archipelago-profile-trees.py $f -o profiles/$(echo $(basename $f) | sed -e 's/\.newick/.profile.csv/g'); done
# for f in trees/*.newick ; do archipelago-summarize.py $f -o summaries/$(echo $(basename $f) | sed -e 's/\.newick/.summary.csv/g'); done
