#! /usr/bin/env python

import sys
import os
import json
import random
import argparse
import itertools
import collections

from archipelago import simulate

"""

## By Island

By habitat-disturbed-vs-undisturbed : 365 taxa, 87 areas, 2 trait states (['0', '1'])
By habitat-generalist-vs-interior   : 198 taxa, 85 areas, 2 trait states (['0', '1'])
By role-widespread-vs-localized     : 365 taxa, 87 areas, 2 trait states (['0', '1'])
By trophic-level                    : 365 taxa, 87 areas, 3 trait states (['0', '1', '2'])

| regime                                | num.tips | root.age   | pure.birth.rate      | trait.rate | area.rate |
| habitat-disturbed-vs-undisturbed      | 365      | 106.998428 | 0.05482309568853121  | 0.062247   | 0.004843  |
| habitat-generalist-vs-interior.newick | 198      | 99.819969  | 0.042834005131106676 | 0.047111   | 0.003863  |
| role-widespread-vs-localized.newick   | 365      | 106.998428 | 0.05482309568853121  | 43.815203  | 0.004843  |
| trophic-level.newick                  | 365      | 106.998428 | 0.05482309568853121  | 0.002914   | 0.004843  |

"""


general_job_template = """\
#! /bin/bash
#$ -cwd
#$ -V
#$ -S /bin/bash
#$ -l h_vmem=12G
#$ -l virtual_free=12G
{queue}
{commands}
"""

def to_json(o):
    if isinstance(o, DynamicValue):
        return o.value
    raise TypeError(repr(o) + " is not JSON serializable")

class DynamicValue(object):

    def __init__(self, value=None):
        self.value = value

def generate_area_definitions(
        num_focal_areas,
        num_supplemental_areas=0):
    areas = []
    for i in range(num_supplemental_areas):
        areas.append({
            "label": "b{}".format(i+1),
            "is_supplemental": True,
            })
    for i in range(num_focal_areas):
        areas.append({
            "label": "a{}".format(i+1),
            "is_supplemental": False,
            })
    return areas

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-replicates-per-job", "--nreps",
            type=int,
            default=100,
            help="Number of simulation replicates (default = %(default)s).")
    parser.add_argument("--num-jobs-per-model", "--njobs",
            type=int,
            default=1,
            help="Number of simulation jobs per model (default = %(default)s).")
    parser.add_argument( "--no-summary",
            action="store_true",
            default=False,
            help="Do not create summaries.")
    parser.add_argument( "--no-profile",
            action="store_true",
            default=False,
            help="Do not create profile.")
    parser.add_argument("--venv",
            default=None,
            help="Path to Python virtual environment.")
    parser.add_argument("-z", "--random-seed",
            default=None,
            help="Seed for random number generator engine.")
    parser.add_argument("-q", "--queue",
            default=None,
            help="Name of queue to use (default: no queue specification")
    args = parser.parse_args()

    if args.random_seed is None:
        args.random_seed = random.randint(0, sys.maxsize)
    rng = random.Random(args.random_seed)
    if args.venv is not None:
        venv_dir = os.path.expanduser(os.path.expandvars(args.venv))
        venv_activate = os.path.abspath(os.path.join(venv_dir, "bin", "activate"))
        if not os.path.exists(venv_activate):
            raise Exception("Virtual environment activation script not found: '{}'".format(venv_activate))
        source_venv = "source {}".format(venv_activate)
    else:
        source_venv = ""
    if args.queue:
        queue = "#$ -q {}".format(args.queue)
    else:
        queue = ""

    lineage_birth_rate = DynamicValue()
    lineage_death_rate = DynamicValue()
    constrained_dispersal_rate = DynamicValue()
    unconstrained_dispersal_rate = DynamicValue()
    target_focal_area_lineages = DynamicValue()

    base_model = {
        "diversification": {
            "lineage_birth_rate": {
                    "definition_type": "fixed_value",
                    "definition": lineage_birth_rate,
            },
            "lineage_death_rate": {
                    "definition_type": "fixed_value",
                    "definition": lineage_death_rate,
            },
        },
        "termination_conditions": {"target_focal_area_lineages": target_focal_area_lineages},
    }
    trait_models = {}
    trait_models["habitat-disturbed-vs-undisturbed"] = {
            "areas": generate_area_definitions(num_focal_areas=87, num_supplemental_areas=1),
            "traits": [
                {"label": "habitat", "nstates": 2, "transition_rate": 0.062247,},
            ],
            "dispersal": {
                "lineage_dispersal_rate": {
                    "definition_type": "trait_state_index_map:habitat",
                    "definition": [unconstrained_dispersal_rate, constrained_dispersal_rate],
                    },
            },
    }
    trait_models["habitat-generalist-vs-interior"] = {
            "areas": generate_area_definitions(num_focal_areas=87, num_supplemental_areas=1),
            "traits": [
                {"label": "habitat", "nstates": 2, "transition_rate": 0.047111,},
            ],
            "dispersal": {
                "lineage_dispersal_rate": {
                    "definition_type": "trait_state_index_map:habitat",
                    "definition": [unconstrained_dispersal_rate, constrained_dispersal_rate],
                    },
            },
    }
    trait_models["role-widespread-vs-localized"] = {
            "traits": [
                {"label": "role_simplified", "nstates": 2, "transition_rate": 43.815203,},
            ],
            "dispersal": {
                "lineage_dispersal_rate": {
                    "definition_type": "trait_state_index_map:role_simplified",
                    "definition": [unconstrained_dispersal_rate, constrained_dispersal_rate],
                    },
            },
    }
    trait_models["trophic-level"] = {
            "traits": [
                {"label": "trophic_level", "nstates": 3, "transition_rate": 0.002914,},
            ],
            "dispersal": {
                "lineage_dispersal_rate": {
                    "definition_type": "trait_state_index_map:trophic_level",
                    "definition": [unconstrained_dispersal_rate, constrained_dispersal_rate, constrained_dispersal_rate],
                    },
            },
    }

    # general_pure_birth_rate = 0.05486829846551938
    # general_extinction_rate = 3.0731e-05
    # general_dispersal_rate = 0.00620009
    trait_birth_death_dispersal_nlineages_regime_map = {
            "habitat-disturbed-vs-undisturbed": (
                (0.05486829846551938, 0.0, 0.004843, 365),
            ),
            "habitat-generalist-vs-interior": (
                (0.042834005131106676, 0.0, 0.003863, 198),
            ),
            "role-widespread-vs-localized": (
                (0.05486829846551938, 0.0, 0.004843, 365),
            ),
            "trophic-level": (
                (0.05486829846551938, 0.0, 0.004843, 365),
            ),
    }

    model_filepaths = []
    job_filepaths = []
    for trait_model_name in trait_models:
        for is_constrained in (True, False):
            param_config_index = 0
            for birth_rate_value, death_rate_value, dispersal_rate_value, target_focal_area_lineages_value in trait_birth_death_dispersal_nlineages_regime_map[trait_model_name]:
                param_config_index += 1
                lineage_birth_rate.value = birth_rate_value
                lineage_death_rate.value = death_rate_value
                target_focal_area_lineages.value = target_focal_area_lineages_value
                if is_constrained:
                    dispersal_model_name = "constrained"
                    constrained_dispersal_rate.value = 0.0
                    unconstrained_dispersal_rate.value = dispersal_rate_value
                else:
                    dispersal_model_name = "unconstrained"
                    constrained_dispersal_rate.value = dispersal_rate_value
                    unconstrained_dispersal_rate.value = dispersal_rate_value
                subdir = trait_model_name
                if not os.path.exists(subdir):
                    os.makedirs(subdir)
                for njob_idx, njob in enumerate(range(args.num_jobs_per_model)):
                    run_name = dispersal_model_name + "_" + trait_model_name + ".{:03d}.{:03d}".format(param_config_index, njob_idx+1)
                    sim_model = {}
                    sim_model.update(base_model)
                    sim_model.update(trait_models[trait_model_name])
                    model_filepath = os.path.join(subdir, run_name + ".json")
                    sys.stderr.write("Model: {}\n".format(model_filepath))
                    with open(model_filepath, "w") as mf:
                        json.dump(sim_model, mf, indent=4, default=to_json)
                    model_filepaths.append(model_filepath)
                    commands = []
                    if source_venv:
                        commands.append(source_venv)
                    run_cmd = []
                    run_cmd.append("archipelago-simulate.py")
                    run_cmd.extend(["--nreps", str(args.num_replicates_per_job)])
                    run_cmd.extend(["-z", str(rng.randint(0, sys.maxsize))])
                    run_cmd.extend(["--output-prefix", run_name])
                    run_cmd.extend(["--stderr-logging-level", "warning"])
                    run_cmd.append(os.path.basename(model_filepath))
                    run_cmd = " ".join(run_cmd)
                    commands.append(run_cmd)
                    result_trees_filepath = simulate.ArchipelagoSimulator.compose_focal_areas_trees_filepath(run_name)
                    if not args.no_profile:
                        profile_cmd = []
                        profile_cmd.append("archipelago-profile-trees.py")
                        profile_cmd.extend(["--model-file", os.path.basename(model_filepath)])
                        # profile_cmd.append("--estimate-dec-biogeobears")
                        profile_cmd.append("--no-estimate-area")
                        profile_cmd.append("--ignore-estimation-errors")
                        profile_cmd.extend(["--verbosity", "0"])
                        profile_cmd.extend(["-o", "{}.profile.csv".format(run_name)])
                        profile_cmd.append(result_trees_filepath)
                        profile_cmd = " ".join(profile_cmd)
                        commands.append(profile_cmd)
                    if not args.no_summary:
                        summary_cmd = []
                        summary_cmd = []
                        summary_cmd.append("archipelago-summarize.py")
                        summary_cmd.extend(["-l", "dispersal.model:{}".format(dispersal_model_name)])
                        summary_cmd.extend(["-o", "{}.summary.csv".format(run_name)])
                        summary_cmd.append("-q")
                        summary_cmd.append(result_trees_filepath)
                        summary_cmd = " ".join(summary_cmd)
                        commands.append(summary_cmd)
                    job_filepath = run_name + ".job"
                    with open(os.path.join(subdir, job_filepath), "w") as jobf:
                        template = general_job_template
                        jobf.write(template.format(
                            commands="\n".join(commands),
                            queue=queue,
                            ))
                    job_filepaths.append((subdir, job_filepath))
    assert len(model_filepaths) == len(set(model_filepaths))
    assert len(job_filepaths) == len(set(job_filepaths))
    with open("submit.sh", "w") as out:
        out.write("#! /bin/bash\n")
        for subdir, job_filepath in job_filepaths:
            out.write("cd {}\n".format(subdir))
            out.write("qsub {}\n".format(job_filepath))
            out.write("cd ..\n")

if __name__ == "__main__":
    main()

