#! /usr/bin/env python

import sys
import os
import json
import random
import argparse
import itertools
import collections
import math

from archipelago import simulate

"""

## By Island

By module, habitat-disturbed-vs-undisturbed : 365 taxa, 4 areas, 2 trait states (['0', '1'])
By module, habitat-generalist-vs-interior   : 198 taxa, 4 areas, 2 trait states (['0', '1'])
By module, role-widespread-vs-localized     : 365 taxa, 4 areas, 2 trait states (['0', '1'])
By module, trophic-level                    : 365 taxa, 4 areas, 3 trait states (['0', '1'])

+----------------------------------+----------+------------+----------------------+-------------+-----------+-----------------------------+------------------------------+
| regime                           | num.tips | root.age   | pure.birth.rate      | trait.rate  | area.rate | lagrange.dec.dispersal.rate | lagrange.dec.extinction.rate |
+----------------------------------+----------+------------+----------------------+-------------+-----------+-----------------------------+------------------------------+
| habitat-disturbed-vs-undisturbed | 365      | 106.998428 | 0.05482309568853121  | 0.062247    | 2.396912  | 0.0179814                   | 0.00257819                   |
| habitat-generalist-vs-interior   | 198      | 99.819969  | 0.042834005131106676 | 0.047111    | 6.566151  | 0.0143178                   | 0.00205547                   |
| role-widespread-vs-localized     | 365      | 106.998428 | 0.05482309568853121  | 43.815203   | 2.396912  | 0.0179814                   | 0.00257819                   |
| module.trophic-level             | 549      | 106.998428 | 0.06286516178810365  | 0.005367171 | 2.396912  | 0.022582881                 | 0.003980697                  |
+----------------------------------+----------+------------+----------------------+-------------+-----------+-----------------------------+------------------------------+

"""


general_job_template = """\
#! /bin/bash
#$ -cwd
#$ -V
#$ -S /bin/bash
#$ -l h_vmem=12G
#$ -l virtual_free=12G
{queue}
{commands}
"""

def to_json(o):
    if isinstance(o, DynamicValue):
        return o.value
    raise TypeError(repr(o) + " is not JSON serializable")

class DynamicValue(object):

    def __init__(self, value=None):
        self.value = value

def generate_area_definitions(
        num_focal_areas,
        num_supplemental_areas=0):
    areas = []
    for i in range(num_supplemental_areas):
        areas.append({
            "label": "b{}".format(i+1),
            "is_supplemental": True,
            })
    for i in range(num_focal_areas):
        areas.append({
            "label": "a{}".format(i+1),
            "is_supplemental": False,
            })
    return areas

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--num-replicates-per-job", "--nreps",
            type=int,
            default=100,
            help="Number of simulation replicates (default = %(default)s).")
    parser.add_argument("--num-jobs-per-model", "--njobs",
            type=int,
            default=1,
            help="Number of simulation jobs per model (default = %(default)s).")

    parser.add_argument("--no-disturbed-vs-undisturbed",
            action="store_true",
            default=False,
            help="Do not create disturbed-vs-undisturbed analysis.")
    parser.add_argument("--no-generalist-vs-interior",
            action="store_true",
            default=False,
            help="Do not create generalist-vs-interior analysis.")
    parser.add_argument("--no-widespread-vs-localized",
            action="store_true",
            default=False,
            help="Do not create widespread-vs-localized analysis.")
    parser.add_argument("--no-trophic-level",
            action="store_true",
            default=False,
            help="Do not create trophic-level analysis.")

    parser.add_argument("--no-summary",
            action="store_true",
            default=False,
            help="Do not create summaries.")
    parser.add_argument("--profile", "--create-profile",
            dest="create_profile",
            action="store_true",
            default=False,
            help="Create profile (estimate birth, death, and dispersal rates on tree).")
    parser.add_argument("--no-profile-dispersal",
            action="store_true",
            default=False,
            help="Do not profile dispersal, even if profile is requested.")
    parser.add_argument("--venv",
            default=None,
            help="Path to Python virtual environment.")
    parser.add_argument("-z", "--random-seed",
            default=None,
            help="Seed for random number generator engine.")
    parser.add_argument("-q", "--queue",
            default=None,
            help="Name of queue to use (default: no queue specification")
    args = parser.parse_args()

    if args.random_seed is None:
        args.random_seed = random.randint(0, sys.maxsize)
    rng = random.Random(args.random_seed)
    if args.venv is not None:
        venv_dir = os.path.expanduser(os.path.expandvars(args.venv))
        venv_activate = os.path.abspath(os.path.join(venv_dir, "bin", "activate"))
        if not os.path.exists(venv_activate):
            raise Exception("Virtual environment activation script not found: '{}'".format(venv_activate))
        source_venv = "source {}".format(venv_activate)
    else:
        source_venv = ""
    if args.queue:
        queue = "#$ -q {}".format(args.queue)
    else:
        queue = ""

    lineage_birth_rate = DynamicValue()
    lineage_death_rate = DynamicValue()
    constrained_dispersal_weight = DynamicValue()
    unconstrained_dispersal_weight = DynamicValue()
    global_dispersal_rate = DynamicValue()
    target_focal_area_lineages = DynamicValue()

    base_model = {
        "diversification": {
            "lineage_birth_rate": {
                    "definition_type": "fixed_value",
                    "definition": lineage_birth_rate,
            },
            "lineage_death_rate": {
                    "definition_type": "fixed_value",
                    "definition": lineage_death_rate,
            },
        },
        "termination_conditions": {"target_focal_area_lineages": target_focal_area_lineages},
    }
    trait_models = {}
    trait_models["habitat-disturbed-vs-undisturbed"] = {
        "traits": [
            {"label": "habitat", "nstates": 2, "transition_rate": 0.062247,},
        ],
        "dispersal": {
            "global_dispersal_rate": global_dispersal_rate,
            "lineage_dispersal_weight": {
                "definition_type": "trait_state_index_map:habitat",
                "definition": [unconstrained_dispersal_weight, constrained_dispersal_weight],
                },
        },
    }
    trait_models["habitat-generalist-vs-interior"] = {
        "traits": [
            {"label": "habitat", "nstates": 2, "transition_rate": 0.047111,},
        ],
        "dispersal": {
            "global_dispersal_rate": global_dispersal_rate,
            "lineage_dispersal_weight": {
                "definition_type": "trait_state_index_map:habitat",
                "definition": [unconstrained_dispersal_weight, constrained_dispersal_weight],
                },
        },
    }
    trait_models["role-widespread-vs-localized"] = {
        "traits": [
            {"label": "role_simplified", "nstates": 2, "transition_rate": 43.815203,},
        ],
        "dispersal": {
            "global_dispersal_rate": global_dispersal_rate,
            "lineage_dispersal_weight": {
                "definition_type": "trait_state_index_map:role_simplified",
                "definition": [unconstrained_dispersal_weight, constrained_dispersal_weight],
                },
        },
    }
    trait_models["trophic-level"] = {
        "traits": [
            {"label": "trophic_level", "nstates": 2, "transition_rate": 0.005367171,},
        ],
        "dispersal": {
            "global_dispersal_rate": global_dispersal_rate,
            "lineage_dispersal_weight": {
                "definition_type": "trait_state_index_map:trophic_level",
                "definition": [unconstrained_dispersal_weight, constrained_dispersal_weight],
                },
        },
    }

    # general_pure_birth_rate = 0.05486829846551938
    # general_extinction_rate = 3.0731e-05
    # general_dispersal_rate = 0.00620009
    trait_birth_death_dispersal_nlineages_regime_map = {}
    if not args.no_disturbed_vs_undisturbed:
        trait_birth_death_dispersal_nlineages_regime_map["habitat-disturbed-vs-undisturbed"] = (
                (0.05486829846551938, 0.00257819, 0.0179814, 365),
                )
    if not args.no_generalist_vs_interior:
        trait_birth_death_dispersal_nlineages_regime_map["habitat-generalist-vs-interior"] = (
                (0.042834005131106676, 0.0, 0.0143178, 198),
            )
    if not args.no_widespread_vs_localized:
        trait_birth_death_dispersal_nlineages_regime_map["role-widespread-vs-localized"] = (
                (0.05486829846551938, 0.00257819, 0.0179814, 365),
            )
    if not args.no_trophic_level:
        trait_birth_death_dispersal_nlineages_regime_map["trophic-level"] = (
                (0.06286516178810365, 0.003980697, 0.022582881, 549),
            )

    model_filepaths = []
    job_filepaths = []
    for trait_model_name in trait_birth_death_dispersal_nlineages_regime_map:
        for is_constrained in (True, False):
            param_config_index = 0
            for num_supplemental_areas_value in (0, 1, 2):
                for birth_rate_value, death_rate_value, global_dispersal_rate_value, target_focal_area_lineages_value in trait_birth_death_dispersal_nlineages_regime_map[trait_model_name]:
                    param_config_index += 1
                    lineage_birth_rate.value = birth_rate_value
                    lineage_death_rate.value = death_rate_value
                    global_dispersal_rate.value = global_dispersal_rate_value
                    target_focal_area_lineages.value = target_focal_area_lineages_value
                    if is_constrained:
                        dispersal_model_name = "constrained"
                        constrained_dispersal_weight.value = 0.0
                        unconstrained_dispersal_weight.value = 1.0 * 2
                    else:
                        dispersal_model_name = "unconstrained"
                        constrained_dispersal_weight.value = 1.0
                        unconstrained_dispersal_weight.value = 1.0
                    subdir = os.path.join("num-supplemental-areas-{}".format(num_supplemental_areas_value), trait_model_name)
                    if not os.path.exists(subdir):
                        os.makedirs(subdir)
                    for njob_idx, njob in enumerate(range(args.num_jobs_per_model)):
                        run_name = dispersal_model_name + "_" + trait_model_name + ".{:03d}.{:03d}".format(param_config_index, njob_idx+1)
                        sim_model = {}
                        sim_model["areas"] = generate_area_definitions(num_focal_areas=4, num_supplemental_areas=num_supplemental_areas_value)
                        sim_model.update(base_model)
                        sim_model.update(trait_models[trait_model_name])
                        model_filepath = os.path.join(subdir, run_name + ".json")
                        sys.stderr.write("Model: {}\n".format(model_filepath))
                        with open(model_filepath, "w") as mf:
                            json.dump(sim_model, mf, indent=4, default=to_json)
                        model_filepaths.append(model_filepath)
                        commands = []
                        if source_venv:
                            commands.append(source_venv)
                        run_cmd = []
                        run_cmd.append("archipelago-simulate.py")
                        run_cmd.extend(["--nreps", str(args.num_replicates_per_job)])
                        run_cmd.extend(["-z", str(rng.randint(0, sys.maxsize))])
                        run_cmd.extend(["--output-prefix", run_name])
                        run_cmd.extend(["--stderr-logging-level", "warning"])
                        run_cmd.append(os.path.basename(model_filepath))
                        run_cmd = " ".join(run_cmd)
                        commands.append(run_cmd)
                        result_trees_filepath = simulate.ArchipelagoSimulator.compose_focal_areas_trees_filepath(run_name)
                        if args.create_profile:
                            profile_cmd = []
                            profile_cmd.append("archipelago-profile-trees.py")
                            profile_cmd.extend(["--model-file", os.path.basename(model_filepath)])
                            if not args.no_profile_dispersal:
                                profile_cmd.append("--estimate-dec-biogeobears")
                            profile_cmd.append("--no-estimate-area")
                            profile_cmd.append("--ignore-estimation-errors")
                            profile_cmd.extend(["--verbosity", "0"])
                            profile_cmd.extend(["-o", "{}.profile.csv".format(run_name)])
                            profile_cmd.append(result_trees_filepath)
                            profile_cmd = " ".join(profile_cmd)
                            commands.append(profile_cmd)
                        if not args.no_summary:
                            summary_cmd = []
                            summary_cmd = []
                            summary_cmd.append("archipelago-summarize.py")
                            summary_cmd.extend(["-l", "dispersal.model:{}".format(dispersal_model_name)])
                            summary_cmd.extend(["-o", "{}.summary.csv".format(run_name)])
                            summary_cmd.append("-q")
                            summary_cmd.append(result_trees_filepath)
                            summary_cmd = " ".join(summary_cmd)
                            commands.append(summary_cmd)
                        job_filepath = run_name + ".job"
                        with open(os.path.join(subdir, job_filepath), "w") as jobf:
                            template = general_job_template
                            jobf.write(template.format(
                                commands="\n".join(commands),
                                queue=queue,
                                ))
                        job_filepaths.append((subdir, job_filepath))
    assert len(model_filepaths) == len(set(model_filepaths))
    assert len(job_filepaths) == len(set(job_filepaths))
    with open("submit.sh", "w") as out:
        out.write("#! /bin/bash\n")
        for subdir, job_filepath in job_filepaths:
            out.write("cd {}\n".format(subdir))
            out.write("qsub {}\n".format(job_filepath))
            out.write("cd -\n")

if __name__ == "__main__":
    main()


"""
notes: equal rates trait matrix converges to a uniform stationary distribution over all states:

> x <- matrix(c(1-0.062247, 0.062247, 0.062247, 1-0.062247),nrow=2)
> y <- x ; for (i in 1:1000) { y <- y %*% x } ; y
     [,1] [,2]
[1,]  0.5  0.5
[2,]  0.5  0.5
> x <- matrix(c(1-0.04, 0.04, 0.04, 1-0.04),nrow=2)
> y <- x ; for (i in 1:1000) { y <- y %*% x } ; y
     [,1] [,2]
[1,]  0.5  0.5
[2,]  0.5  0.5
"""
