#! /usr/bin/env python
# -*- coding: utf-8 -*-

import collections
import csv
import dendropy

class SystemConfiguration(object):

    def __init__(
            self,
            path_to_master_reference,
            path_to_master_tree,
            ):
        self.load_master_reference(path_to_master_reference)
        self.master_tree_path = path_to_master_tree
        self.master_tree_schema = "nexus"

    def load_master_reference(self, path):
        self.master_reference_path = path
        self.lineage_names = []
        self.lineage_name_disturbed_vs_undisturbed_habitat_map = collections.OrderedDict()
        self.lineage_name_generalist_vs_interior_habitat_map = collections.OrderedDict()
        self.lineage_name_widespread_vs_localized_role_map = collections.OrderedDict()
        self.lineage_name_trophic_level_map = collections.OrderedDict()
        self.island_lineage_names_map = {}
        self.lineage_name_islands_map = {}
        self.module_lineage_names_map = {}
        self.lineage_name_modules_map = {}
        self.jetz_to_study_name_map = {}
        reader = csv.DictReader(open(path, "r"))
        fieldnames = reader.fieldnames
        self.islands = sorted([fieldname for fieldname in fieldnames if fieldname.startswith("island_area")])
        self.modules = sorted([fieldname for fieldname in fieldnames if fieldname.startswith("module_area")])
        for row in reader:

            lineage_name = row["operational_name"]
            self.jetz_to_study_name_map[row["jetz_name"]] = lineage_name

            if "name mapping failure" in row["study_status"]:
                continue

            assert lineage_name
            self.lineage_names.append(lineage_name)

            if "non-informative habitat category" not in row["study_status"] and "no habitat information" not in row["study_status"]:
                habitat = row["habitat"]
                assert habitat, "{}: '{}' ({})".format(lineage_name, habitat, row["study_status"])
                if habitat == "1":
                    self.lineage_name_disturbed_vs_undisturbed_habitat_map[lineage_name] = "1"
                    self.lineage_name_generalist_vs_interior_habitat_map[lineage_name] = "1"
                else:
                    assert habitat in ("2", "3", "4", "5")
                    self.lineage_name_disturbed_vs_undisturbed_habitat_map[lineage_name] = "0"
                    if habitat == "5":
                        self.lineage_name_generalist_vs_interior_habitat_map[lineage_name] = "0"
                    else:
                        self.lineage_name_generalist_vs_interior_habitat_map[lineage_name] = None
            else:
                self.lineage_name_generalist_vs_interior_habitat_map[lineage_name] = None
                self.lineage_name_disturbed_vs_undisturbed_habitat_map[lineage_name] = None

            role = row["role"]
            if role == "Peripheral":
                self.lineage_name_widespread_vs_localized_role_map[lineage_name] = "1"
            else:
                assert role in ("Connector", "Module hub", "Network hub")
                self.lineage_name_widespread_vs_localized_role_map[lineage_name] = "0"

            if "no guild information" not in row["study_status"]:
                # Frugivores # Granivores # Herbivores # Nectarivores # Insectivores # Invertabrates # Omnivores # Piscivore # Carnivore
                guild_name = row["guild"]
                if guild_name in ("Frugivores", "Granivores", "Herbivores", "Nectarivores", "Omnivores"):
                    trophic_level = "0"
                elif guild_name in ("Insectivores", "Invertabrates", "Piscivore", "Carnivore"):
                    trophic_level = "1"
                else:
                    raise ValueError(guild_name)
                self.lineage_name_trophic_level_map[lineage_name] = trophic_level
            else:
                self.lineage_name_trophic_level_map[lineage_name] = None

            for island in self.islands:
                assert row[island] in ("0", "1")
                if row[island] == "1":
                    try:
                        self.island_lineage_names_map[island].add(lineage_name)
                    except KeyError:
                        self.island_lineage_names_map[island] = set([lineage_name])
                    try:
                        self.lineage_name_islands_map[lineage_name].add(island)
                    except KeyError:
                        self.lineage_name_islands_map[lineage_name] = set([island])

            for module in self.modules:
                assert row[module] in ("0", "1")
                if row[module] == "1":
                    try:
                        self.module_lineage_names_map[module].add(lineage_name)
                    except KeyError:
                        self.module_lineage_names_map[module] = set([lineage_name])
                    try:
                        self.lineage_name_modules_map[lineage_name].add(module)
                    except KeyError:
                        self.lineage_name_modules_map[lineage_name] = set([module])

        for island in list(self.islands):
            if island not in self.island_lineage_names_map:
                self.islands.remove(island)

        for module in list(self.modules):
            if module not in self.module_lineage_names_map:
                self.modules.remove(module)

        self.study_lineages = set(self.lineage_names)

    def get_new_master_tree(self,
            restrict_to_study_lineages=True,
            additional_lineage_names_to_drop=None,
            distribution_vector_type=None,
            ):
        tree = dendropy.Tree.get_from_path(
                self.master_tree_path,
                self.master_tree_schema,
                )
        for taxon in tree.taxon_namespace:
            taxon.label = self.jetz_to_study_name_map[taxon.label]

        lineage_names_to_drop = set()
        if restrict_to_study_lineages:
            for taxon in tree.taxon_namespace:
                if taxon.label not in self.study_lineages:
                    lineage_names_to_drop.add(taxon.label)
        if additional_lineage_names_to_drop:
            lineage_names_to_drop.update(additional_lineage_names_to_drop)
        tree.prune_taxa_with_labels(lineage_names_to_drop)
        tree.purge_taxon_namespace()

        if distribution_vector_type is not None:
            if distribution_vector_type == "island":
                island_taxa = {}
                for taxon in tree.taxon_namespace:
                    for island in self.islands:
                        if island in self.lineage_name_islands_map[taxon.label]:
                            try:
                                island_taxa[island].add(taxon)
                            except KeyError:
                                island_taxa[island] = set([taxon])
                for taxon in tree.taxon_namespace:
                    v = []
                    for island in list(self.islands):
                        if island in island_taxa:
                            if taxon in island_taxa[island]:
                                v.append("1")
                            else:
                                v.append("0")
                    assert "1" in v
                    taxon.distribution_vector = v
                num_islands = len(tree.taxon_namespace[0].distribution_vector)
                for taxon in tree.taxon_namespace:
                    assert len(taxon.distribution_vector) == num_islands
                tree.num_areas = num_islands
            elif distribution_vector_type == "module":
                module_taxa = {}
                for taxon in tree.taxon_namespace:
                    for module in self.modules:
                        if module in self.lineage_name_modules_map[taxon.label]:
                            try:
                                module_taxa[module].add(taxon)
                            except KeyError:
                                module_taxa[module] = set([taxon])
                for taxon in tree.taxon_namespace:
                    v = []
                    for module in list(self.modules):
                        if module in module_taxa:
                            if taxon in module_taxa[module]:
                                v.append("1")
                            else:
                                v.append("0")
                    assert "1" in v
                    taxon.distribution_vector = v
                num_modules = len(tree.taxon_namespace[0].distribution_vector)
                for taxon in tree.taxon_namespace:
                    assert len(taxon.distribution_vector) == num_modules
                tree.num_areas = num_modules
            else:
                raise ValueError(distribution_vector_type)

        return tree




